#!/usr/bin/env python
#coding=utf-8

# $Id: pymplayer.py 435 2008-02-09 09:59:11Z darwin $
#
# Copyright (C) 2007-2008  UP EEE Computer Networks Laboratory
# Copyright (C) 2007-2008  Darwin M. Bautista <djclue917@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Thin, out-of-source wrapper and client/server for MPlayer

Classes:

MPlayer -- thin, out-of-process wrapper for MPlayer
Server -- asynchronous server that manages an MPlayer instance
Client -- client for sending MPlayer commands

Function:

loop() -- the asyncore.loop function, for use in conjunction with Client

Constants:

PORT -- default port used by Client and Server
MAX_CMD_LEN -- maximum length of a command

"""

import time
import thread
from subprocess import Popen, PIPE
from system import *
import os
if os =="nt":
    _mplayerext=".exe"
    _creationflag=0x8000000
else:
    _mplayerext=""
    _creationflag=0
__all__ = [
    'MPlayer',
    'loop',
    'MAX_CMD_LEN'
    ]

__version__ = '0.3.0'
__author__ = 'Darwin M. Bautista <djclue917@gmail.com>'


PORT = 50001
MAX_CMD_LEN = 256

def _runcmd(cmd):
    return subprocess.Popen(cmd,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE,universal_newlines=True,creationflags=_creationflag)
class videofileinfo(object):
    def __init__(self,mplayercmd=None):
        if mplayercmd is None:
            self.mplayercmd="mplayer"+_mplayerext
        else:
            self.mplayercmd=mplayercmd
    def getfileinfo(self,filename):
        self.length=0
        self.width=0
        self.height=0
        _getfileinfo=_runcmd([self.mplayercmd,filename,"-vo","null","-vc", "dummy","-ac","dummy","-identify"])
        _getfileinfo.wait()
        _fileinfo=_getfileinfo.stdout.readlines()
        _info={}
        for line in _fileinfo:
            _pos=line.find("=")
            if _pos>0:
                _info[line[:_pos]]=line[_pos+1:].strip()
        if _info.has_key("ID_LENGTH"):
            self.length=float(_info["ID_LENGTH"])
        if _info.has_key("ID_VIDEO_WIDTH"):
            self.width=int(_info["ID_VIDEO_WIDTH"])
        if _info.has_key("ID_VIDEO_HEIGHT"):
            self.height=int(_info["ID_VIDEO_HEIGHT"])
        if self.width!=0:
            self.poshint="A:"
            self.poshint2="V:"
            self.poshint3="A-V:"
        else:
            self.poshint="A:"
            self.poshint2="A:"
            self.poshint3="("
class MPlayer(videofileinfo):
    """MPlayer(args=())

    An out-of-process wrapper for MPlayer. It provides the basic interface
    for sending commands and receiving responses to and from MPlayer. Take
    note that MPlayer is always started in 'slave', 'idle', and 'quiet' modes.

    The MPlayer process would eventually "freeze" if the poll_output method
    is not called because the stdout/stderr PIPE buffers would get full.
    Also, the handle_data and handle_error methods would only get called,
    given an I/O event, after the poll_output method is called.

    A different 'I/O watcher' can be used by overriding the create_handler
    and remove_handler methods.

    @class attribute mplayercmd: path to or filename of the MPlayer executable
    @property args: MPlayer arguments

    """
    length=0
    width=0
    height=0
    def __init__(self, args=()):
        self.args = args
        self.mplayercmd="mplayer"+_mplayerext
        self.wid=None
        self._map = {}
        self._process = None
        self._response = None
        
    def __del__(self):
        # Be sure to stop the MPlayer process.
        self.stop()

    def _get_args(self):
        return self._args[3:]

    def _set_args(self, args):
        if not isinstance(args, (list, tuple)):
            raise TypeError("args should either be a tuple or list of strings")
        elif args:
            for arg in args:
                if not isinstance(arg, basestring):
                    raise TypeError("args should either be a tuple or list of strings")
        self._args = ['-slave','-idle','-noquiet',"-nomouseinput"]
        self._args.extend(args)

    args = property(_get_args, _set_args, doc="MPlayer arguments")

    def _handle_data(self,data):
        if data.startswith(self.poshint):
            self.handle_pos(float(data[data.find(self.poshint2)+2:data.find(self.poshint3)]))
        else:
            self.handle_data(data)
    def handle_pos(self,pos):
        return
    def start(self,filename=None):
        """Start the MPlayer process.

        Returns True on success, False on failure,
        and None if MPlayer is already running.

        """
        if not self.isalive():
            self.paused=False
            args = [self.mplayercmd]
            if self.wid!=None:
                args.extend(["-wid",str(self.wid)])
            if filename!=None:
                args.extend([filename])
            args.extend(self._args)
            print args
            if True:
            #try:
                # Start the MPlayer process (line-buffered)
                self.getfileinfo(filename)
                self._process =_runcmd(args)
            #except OSError:
                #return False
            #else:
                self.stdoutwatch=pipewatch(self._process,self._process.stdout, self._handle_data)
                self.stderrwatch=pipewatch(self._process,self._process.stderr, self.handle_error)
                return True

    def stop(self):
        """Stop the MPlayer process.

        Returns the exit status of MPlayer or None if not running.

        """
        if self.isalive():
            self.command("quit")
            self.stdoutwatch.stop()
            self.stderrwatch.stop()
            #self.remove_handler(self._process.stderr)
            return self._process.wait()

    def restart(self):
        """Convenience method for restarting the MPlayer process.

        Restarting MPlayer means stopping the current process and
        starting a new one which means that the poll_output method
        will finish and return and will need to be called again.

        Returns the return values of the stop and start methods as a 2-tuple.

        """
        return self.stop(), self.start()

    def command(self, cmd, timeout=0.1):
        """Send a command to MPlayer.

        @param cmd: command string
        @param timeout: time to wait before returning command output

        Returns the output if command is a valid get_* command.
        Else, None is returned.

        Valid MPlayer commands are documented in:
        http://www.mplayerhq.hu/DOCS/tech/slave.txt

        """
        if not isinstance(cmd, basestring):
            raise TypeError("command must be a string")
        if self.isalive() and cmd:
            self._process.stdin.write("".join([cmd, '\n']))
            self._process.stdin.flush()

    def isalive(self):
        """Check if MPlayer process is alive.

        Returns True if alive, else, returns False.

        """
        try:
            return (self._process.poll() is None)
        except AttributeError:
            return False

    @staticmethod
    def handle_data(data):
        """Handle the data read from stdout.

        This method is meant to be overridden.
        It will be called for each readline() from stdout.

        @param data: the line read from stdout

        """

    @staticmethod
    def handle_error(data):
        """Handle the data read from stderr.

        This method is meant to be overridden.
        It will be called for each readline() from stderr.

        @param data: the line read from stderr

        """
        return True

    def toggleplay(self):
        self.command("pause")
        self.paused=not self.paused
    def psend(self,str):
        if self.paused:
            str="pausing "+str
        self.command(str)
    def setpos(self,pos):
        self.psend("set_property time_pos "+str(pos))
class pipewatch:
    def __init__(self,pipe,file,callback):
        self.stoped=False
        self.pipe=pipe
        self.file=file
        self.callback=callback
        self.thread=thread.start_new_thread(self.watch,(pipe,file))
    def watch(self,pipe,file):
        while (not self.stoped) and pipe.poll()==None:
            data=file.readline()
            file.flush()
            if data:
                try:
                    self.callback(data.rstrip())
                except:
                    pass
        self.stop=True
    def stop(self):
        self.stoped=True
